package snakes;
import ij.*;
import ij.gui.*;
import ij.process.*;
import java.util.*;
import Jama.*;

/**
 * This class models an active contour as a polygonal line with a fixed number of vertices.
 * The method 'iteration' corresponds to one iteration of the deformation.
 */
public class Snake {

	private double[] X; // vector of the abscissae of the control points
	private double[] Y; // vector of the ordinates of the control points
	private ImageProcessor image;
	private FloatProcessor SGImage; // norm of the squared gradient of the image
	private double alpha, beta, gamma, lambda; // coefficients of the forces (stretching, bending, viscosity)
	private int N = 0;
	private Matrix M;
	
	/**
	 * Computes the squared norm of the gradient of [ip].
	 *
	 * @param ip gray level image.
	 * @return edge image.
	 */
	static FloatProcessor squaredGradientImage(ImageProcessor ip){
		int w = ip.getWidth();
		int h = ip.getHeight();
		FloatProcessor sg = new FloatProcessor(w, h);
		FloatProcessor gx = new FloatProcessor(w, h);
		FloatProcessor gy = new FloatProcessor(w, h);

		ByteProcessor gn = 
			Deriche.deriche( ip, 0.5, gx, gy );
		for(int x=0; x<w; x++)
			for(int y=0; y<h; y++){
				float xg = gx.getf(x, y);
				float yg = gy.getf(x, y);
				sg.setf(x, y, xg*xg+yg*yg);
			}
		return sg;
	}

	/**
	 * Constructor. Gives an invalid snake.
	 * @see initRectangle,initParameters
	 */
	public Snake(){
		N = 0;
	}


	/**
	 * Initializes the snake as a rectangle with [nb] points by
	 * side. 
	 *
	 * @param x1 abscissa of the top-left point.
	 * @param y1 ordinate of the top-left point.
	 * @param x2 abscissa of the bottom-right point.
	 * @param y2 ordinate of the bottom-right point.
	 * @param nb number of points on one side.
	 */
	public void initRectangle(double x1, double y1, double x2, double y2, int nb){
		N = 4*nb;
		X = new double[N];
		Y = new double[N];
		int j = 0;

		for(int i=0; i <=nb; i++){
			X[j] = x1;
			Y[j] = y1+((y2-y1)*i)/nb;
			j++;
		}
		for(int i=1; i <=nb; i++){
			X[j] = x1+((x2-x1)*i)/nb;	
			Y[j] = y2;
			j++;
		}

		for(int i=1; i <=nb; i++){
			X[j] = x2;
			Y[j] = y2-((y2-y1)*i)/nb;
			j++;
		}
		for(int i=1; i <nb; i++){
			X[j] = x2-((x2-x1)*i)/nb;
			Y[j] = y1;
			j++;
		}
	}

	/**
	 * Initializes the snake parameters.
	 *
	 * @param ip image where we extract a region
	 * according to the edge information.
	 *
	 * @param alpha elasticity coefficient
	 * @param beta rigidity coefficient
	 * @param gamma friction coefficient
	 * @param lambda influence of the image forces.
	 */
	public void initParameters(ImageProcessor ip, double alpha, double beta, double gamma, double lambda ){
		if ( N == 0 ) throw new RuntimeException( "Snake without geometry.");

		this.image = ip;
		this.alpha = alpha;
		this.beta = beta;
		this.gamma = gamma;
		this.lambda = lambda;
		SGImage = squaredGradientImage(image);
		// TODO : compute M
	}


	/**
	 * @return number of points of the snake.
	 */
	public int getNbPts(){
		return N;
	}

	/**
	 * @param i index between 0 and 'getNbPts()'.
	 * @return coordinate X[i].
	 */
	public double getX(int i){
		return X[i];
	}

	/**
	 * @param i index between 0 and 'getNbPts()'.
	 * @return coordinate Y[i].
	 */
	public double getY(int i){
		return Y[i];
	}

	/**
	 * @return segmented image.
	 */
	public ImageProcessor getImage(){
		return image;
	}

	/**
	 * @return contour image.
	 */
	public FloatProcessor getSGImage(){
		return SGImage;
	}

	/**
	 * Compute the product matrix [A] by vector [V] (sizes should be 
	 * compatible).
	 *
	 * @param A matrix with m rows and n columns
	 * @param V vector of size n (n rows)
	 * @return vector of size m (m rows).
	 */
	private double[] mult(Matrix A, double[] V){
		int n = V.length;
		int m = A.getRowDimension();
		double[] res = new double[m];
		for(int i=0; i<m; i++){
			res[i] = 0.0;
			for(int j=0; j<n; j++)
				res[i] += A.get(i, j)*V[j];
		}
		return res;
	}

	
	/**
	 * Compute the vectors 'X' et 'Y' at time step t+1 from
	 * the location of the snake at time t.
	 */
	public void iteration(){
	    // TODO
	}

}

